import matplotlib.pyplot as pt
import numpy as np
import qcodes as qc
from q1simulator import Cluster as SimCluster
from qblox_instruments import Cluster

def generate_waveforms():
    ''' Generates ramps '''
    waveforms = {}
    for i, (start, stop) in enumerate([
            (-1.0, -0.5), (-0.5, 0.0), (0.0, 0.5), (0.5, 1.0),
            (1.0, 0.5), (0.5, 0.0), (0.0, -0.5), (-0.5, -1.0),
            ]):
        waveforms[f'ramp{i}'] = {
                'data': (0.5*np.linspace(start, stop, 60)).tolist(),
                'index': i,
                }
    return waveforms


class TestCluster:

    def __init__(self, sim=True):
        qc.Instrument.close_all()
        self.sim = sim
        if sim:
            cluster = SimCluster('test', {1:'QCM', 2:'QRM'})
            qcm = cluster.module1
            # qcm.config('trace', True)
            qrm = cluster.module2
            # qrm.config('trace', True)
            cluster.config('render_repetitions', False)
            cluster.config('skip_wait_sync', True)
        else:
            cluster = Cluster('Qblox_Cluster', '192.168.0.3')
            qcm = cluster.module2
            qrm = cluster.module10

        self.cluster = cluster
        self.qcm = qcm
        self.qrm = qrm
        self.reset()

    def reset(self):
        self.cluster.reset()
        self.armed = []
        for module in [self.qcm, self.qrm]:
            seq = module.sequencers[0]
            seq.connect_out0('I')
            seq.label = f"{module.name}-0"
            seq = module.sequencers[1]
            seq.connect_out1('Q')
            seq.label = f"{module.name}-1"
        for num in [0,1]:
            seq = self.qrm.sequencers[num]
            seq.thresholded_acq_trigger_en(False)
            seq.integration_length_acq(100)


    def load(self, on_qcm, sequencer_number, program, waveforms={}, weights={}, acquisitions={}):
        if on_qcm:
            module = self.qcm
        else:
            module = self.qrm
        seq = module.sequencers[sequencer_number]
        seq.sequence({
                'program': program,
                'waveforms': waveforms,
                'weights': weights,
                'acquisitions': acquisitions})
        seq.sync_en(True)
        module.arm_sequencer(sequencer_number)
        self.armed.append([module.slot_idx, sequencer_number])

    def trigger_out(self, sequencer_number, address, threshold:float,
                    invert:bool=False):
        seq = self.qrm.sequencers[sequencer_number]
        seq.thresholded_acq_trigger_en(True)
        seq.thresholded_acq_trigger_address(address)
        seq.thresholded_acq_threshold(threshold)
        seq.thresholded_acq_trigger_invert(invert)

    def set_trigger_thresholding(self, on_qcm, sequencer_number,
                                 address: int, count: int = 1,
                                 invert: bool = False):
        if on_qcm:
            module = self.qcm
        else:
            module = self.qrm
        seq = module.sequencers[sequencer_number]
        seq.set_trigger_thresholding(address, count, invert)

    def run(self):
        self.cluster.start_sequencer()
        if self.sim:
            pt.figure()
            self.qcm.plot()
            self.qrm.plot()
            pt.legend()
            pt.grid(True)

def qcm_program(used_triggers):
    mask = 0
    for addr in used_triggers:
        mask |= 1 << (addr - 1)

    return f'''
    move {n_rep},R1
    wait_sync 100
    _start:
      set_latch_en 1, 100
      set_awg_gain 32767,32767
      set_mrk 1
      play 0,0,100
      set_mrk 0
      play 3,3,100
      wait 92 # subtract 8 for next conditional

      # set_cond enable, mask, OR, else_wait
      set_cond 1, {mask}, 0, 4
      wait 8 # 4ns * number of statements including this wait
      play 1,1,92 # subtract 8 ns for next conditional
      # duration of else: 8 ns

      # set_cond enable, 0b0001, NOR, else_wait
      set_cond 1, {mask}, 1, 4
      play 6,6,80
      wait 20

      # disable conditional
      set_cond 0, 0, 0, 0
      play 3,3,96
      latch_rst 4
    loop R1,@_start
    stop
    '''

def qrm_program(used_triggers):
    mask = 0
    for addr in used_triggers:
        mask |= 1 << (addr - 1)

    return f'''
    move {n_rep},R1
    wait_sync 100
    _start:
      set_latch_en 1, 4
      acquire 0, 0, 96
      set_awg_gain 32767,32767
      play 0,0,100
      play 3,3,100
      wait 92 # subtract 8 for next conditional

      # set_cond enable, mask, OR, else_wait
      set_cond 1, {mask}, 0, 4
      wait 8 # 4ns * number of statements including this wait
      play 1,1,92 # subtract 8 ns for next conditional
      # duration of else: 8 ns

      # set_cond enable, 0b0001, NOR, else_wait
      set_cond 1, {mask}, 1, 4
      play 6,6,80
      wait 20

      # disable conditional
      set_cond 0, 0, 0, 0
      play 3,3,96
      latch_rst 4
    loop R1,@_start
    stop
    '''

sim = TestCluster(sim=True)
waveforms = generate_waveforms()
acquisitions = {
        "acq0": {"num_bins": 1, "index": 0},
        }

n_rep = 20 #_000_000


# %%

# qcm0: uses [1,2], qrm0: emits 1, qrm1 emits 2,

sim.reset()
sim.load(True, 0, qcm_program([1, 2]), waveforms=waveforms)
sim.load(False, 0, qrm_program([]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(0, 1, 0.0)
sim.load(False, 1, qrm_program([]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(1, 2, 0.0)
sim.cluster.config('acq_trigger_value', 0)
sim.run()
sim.cluster.config('acq_trigger_value', 1)
sim.run()

# %%

# qcm0: uses 1, qrm1: emits 1 uses 2, qrm2 emits 2

sim.reset()
sim.load(True, 0, qcm_program([1]), waveforms=waveforms)
sim.load(False, 0, qrm_program([2]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(0, 1, 0.0)
sim.load(False, 1, qrm_program([]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(1, 2, 0.0)
sim.cluster.config('acq_trigger_value', 0)
sim.run()
sim.cluster.config('acq_trigger_value', 1)
sim.run()

# %%

# qcm0: uses 1, qrm1: emits 1 uses 2, qrm2 emits 2, uses 2

sim.reset()
sim.load(True, 0, qcm_program([1]), waveforms=waveforms)
sim.load(False, 0, qrm_program([2]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(0, 1, 0.0)
sim.load(False, 1, qrm_program([2]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(1, 2, 0.0)
sim.cluster.config('acq_trigger_value', 0)
sim.run()
sim.cluster.config('acq_trigger_value', 1)
sim.run()

# %%

# qcm0: uses 2 qrm1: emits 1 uses 1, 2, qrm2 emits 2, uses 2

sim.reset()
sim.load(True, 0, qcm_program([2]), waveforms=waveforms)
sim.load(False, 0, qrm_program([1, 2]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(0, 1, 0.0)
sim.load(False, 1, qrm_program([2]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(1, 2, 0.0)
sim.cluster.config('acq_trigger_value', 0)
sim.run()
sim.cluster.config('acq_trigger_value', 1)
sim.run()

# %%

# Dependencies on not configured triggers

sim.reset()
sim.load(True, 0, qcm_program([1, 2, 3]), waveforms=waveforms)
sim.load(False, 0, qrm_program([1, 2]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(0, 1, 0.0)
sim.load(False, 1, qrm_program([2, 9]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(1, 2, 0.0)
sim.cluster.config('acq_trigger_value', 0)
sim.run()
sim.cluster.config('acq_trigger_value', 1)
sim.run()


# %%

# qcm0: uses 1,2 qrm1: uses 1, 2, qrm2 uses 2

# !!! SORTING FAILS !!!
sim.reset()
sim.load(True, 0, qcm_program([]), waveforms=waveforms)
sim.load(False, 0, qrm_program([2]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(0, 1, 0.0)
sim.load(False, 1, qrm_program([1, 2]), acquisitions=acquisitions, waveforms=waveforms)
sim.trigger_out(1, 2, 0.0)
try:
    sim.run()
except Exception as ex:
    if ex.args[0].startswith("Sorting sequencers on triggers failed."):
        print("Test OK.")
    else:
        raise





from q1simulator import Q1Simulator

sim = Q1Simulator("test", 1, "QCM")

#%%
seq = sim.sequencer0
seq.connect_out0("I")
seq.sync_en(True)
seq.gain_awg_path0(-0.5)
seq.offset_awg_path0(0.1)

sequence = {
    "waveforms": {
        "ones80": {
            "data": [1.0]*80,
            "index": 0
            },
        },
    "weights": {},
    "acquisitions": {},
    "program": """
       wait 20
       play 0, 0, 100
       set_awg_offs 32767, 32767
       upd_param 100
       set_awg_offs 0,0
       upd_param 100
       stop
    """,
    }

seq.sequence(sequence)

sim.arm_sequencer(0)
sim.start_sequencer(0)
sim.get_sequencer_status(0)

sim.plot()



from qblox_instruments import __version__ as qblox_version_str
from packaging.version import Version
from q1simulator import __version__ as q1simulator_version


def check_qblox_instrument_version():
    min_version = "0.12.0"
    max_version = "0.18.0"
    if qblox_version < Version(min_version):
        raise Exception(f'Q1Simulator {q1simulator_version} requires qblox_instruments v{min_version}+')
    if qblox_version > Version(max_version):
        print(f'WARNING Q1Simulator {q1simulator_version} simulates up to qblox_instruments v{max_version}, '
              f'but qblox_instruments version {qblox_version} is installed')


qblox_version = Version(qblox_version_str)



import logging
from functools import partial

import numpy as np
import matplotlib.pyplot as pt
import qcodes as qc

from .q1sequencer import Q1Sequencer
from .qblox_version import check_qblox_instrument_version, qblox_version, Version
from .triggers import TriggerDistributor
from .trigger_sorting import get_seq_trigger_info, sort_sequencers

from qblox_instruments import (
        InstrumentClass, InstrumentType,
        SystemStatuses, SystemStatus, SystemStatusSlotFlags,
        )

if qblox_version < Version('0.14'):
    from qblox_instruments import SystemState, SystemStatusOld


logger = logging.getLogger(__name__)


class Q1Module(qc.instrument.InstrumentBase):
    _module_parameters = [
        'ext_trigger_input_delay',
        'ext_trigger_input_trigger_en',
        'ext_trigger_input_trigger_address',
        'marker0_inv_en',
        'marker1_inv_en',
        'marker2_inv_en',
        'marker3_inv_en',
        ]
    _sim_parameters_qcm = [
        'out0_offset',
        'out1_offset',
        'out2_offset',
        'out3_offset',
        ]
    _sim_parameters_qcm_rf = [
        'out0_lo_freq',
        'out1_lo_freq',
        'out0_lo_en',
        'out1_lo_en',
        'out0_att',
        'out1_att',
        'out0_offset_path0',
        'out0_offset_path1',
        'out1_offset_path0',
        'out1_offset_path1',
        'out0_lo_freq_cal_type_default',
        'out1_lo_freq_cal_type_default',
        ]
    _sim_parameters_qrm = [
        'out0_offset',
        'out1_offset',
        'in0_gain',
        'in1_gain',
        'scope_acq_trigger_mode_path0',
        'scope_acq_trigger_mode_path1',
        'scope_acq_trigger_level_path0',
        'scope_acq_trigger_level_path1',
        'scope_acq_sequencer_select',
        'scope_acq_avg_mode_en_path0',
        'scope_acq_avg_mode_en_path1',
        'in0_offset',
        'in1_offset',
        ]
    _sim_parameters_qrm_rf = [
        'in0_att',
        'out0_att',
        'in0_offset_path0',
        'in0_offset_path1',
        'out0_in0_lo_freq',
        'out0_in0_lo_en',
        'out0_offset_path0',
        'out0_offset_path1',
        'scope_acq_trigger_mode_path0',
        'scope_acq_trigger_mode_path1',
        'scope_acq_trigger_level_path0',
        'scope_acq_trigger_level_path1',
        'scope_acq_sequencer_select',
        'scope_acq_avg_mode_en_path0',
        'scope_acq_avg_mode_en_path1',
        'out0_in0_lo_freq_cal_type_default',
        ]

    # NOTE: No __init__() !!!
    # This class is used as a mixin. Although quite heavy mixin.

    def init_module(self, n_sequencers=6, sim_type=None):
        self._sim_type = sim_type
        if sim_type is None:
            raise Exception('sim_type must be specified')

        self._is_qcm = sim_type in ['QCM', 'QCM-RF', 'Viewer']
        self._is_qrm = sim_type in ['QRM', 'QRM-RF', 'Viewer']
        self._is_rf = sim_type in ['QCM-RF', 'QRM-RF']

        if not (self._is_qcm or self._is_qrm):
            raise ValueError(f'Unknown sim_type: {sim_type}')

        sim_params = []
        sim_params += self._module_parameters
        if sim_type == 'QCM':
            sim_params += self._sim_parameters_qcm
        elif sim_type == 'QCM-RF':
            sim_params += self._sim_parameters_qcm_rf
        elif sim_type == 'QRM':
            sim_params += self._sim_parameters_qrm
        elif sim_type == 'QRM-RF':
            sim_params = self._sim_parameters_qrm_rf
        elif sim_type == 'Viewer':
            sim_params += list(set(self._sim_parameters_qcm+self._sim_parameters_qrm))

        n_out_ch_type = {
            'QCM': 4,
            'QRM': 2,
            'QCM-RF': 2,
            'QRM-RF': 1,
            'Viewer': 0,
            }
        for ch in range(n_out_ch_type[sim_type]):
            sim_params += [
                f'out{ch}_latency',
                f'out{ch}_fir_config',
                f'out{ch}_fir_coeffs',
                ]
            sim_params += [f'out{ch}_exp{i}_config' for i in range(4)]
            if sim_type == 'QCM':
                sim_params += [f'out{ch}_exp{i}_time_constant' for i in range(4)]
                sim_params += [f'out{ch}_exp{i}_amplitude' for i in range(4)]

        for ch in range(4):
            sim_params += [
                f'marker{ch}_fir_config',
                ]
            sim_params += [f'marker{ch}_exp{i}_config' for i in range(4)]

        for par_name in sim_params:
            self.add_parameter(par_name, set_cmd=partial(self._set, par_name))

        self.sequencers = [Q1Sequencer(self, f'seq{i}', sim_type)
                           for i in range(n_sequencers)]
        for i, seq in enumerate(self.sequencers):
            self.add_submodule(f'sequencer{i}', seq)

        if sim_type == 'QCM-RF':
            self.out0_lo_cal = lambda: logger.info("Calibrate LO 0")
            self.out1_lo_cal = lambda: logger.info("Calibrate LO 1")
        if sim_type == 'QRM-RF':
            self.out0_in0_lo_cal = lambda: logger.info("Calibrate LO 0")

        self.armed_seq = set()
        if self._is_qrm:
            if self._is_rf:
                self.in0_att(0)
            else:
                self.in0_gain(0)
                self.in1_gain(0)

    @property
    def module_type(self) -> InstrumentType:
        return InstrumentType.QCM if self._is_qcm else InstrumentType.QRM

    @property
    def is_qcm_type(self):
        return self._is_qcm

    @property
    def is_qrm_type(self):
        return self._is_qrm

    @property
    def is_rf_type(self):
        return self._is_rf

    def reset(self):
        self.armed_seq = set()
        for seq in self.sequencers:
            seq.reset()

    def _set(self, name, value):
        logger.info(f'{self.name}:{name}={value}')

    def _seq_set(self, name, value):
        seq_nr = int(name[9])
        self.sequencers[seq_nr]._set_legacy(name[11:], value)

    def get_num_system_error(self):
        return 0

    def get_system_error(self):
        return '0,"No error"'

    def arm_sequencer(self, seq_nr):
        self.armed_seq.add(seq_nr)
        self.sequencers[seq_nr].arm()

    def start_sequencer(self, sequencer: int | None = None):
        start_indices = self.armed_seq if sequencer is None else (sequencer,)
        for seq_nr in start_indices:
            self.sequencers[seq_nr].run()

    def stop_sequencer(self, sequencer: int | None = None):
        self.armed_seq = set()

    # Deprecated API
    if qblox_version < Version('0.14'):
        def get_sequencer_state(self, seq_nr, timeout=0):
            return self.sequencers[seq_nr].get_state()

        def get_acquisition_state(self, seq_nr, timeout=0):
            return self.sequencers[seq_nr].get_acquisition_state()

    def get_sequencer_status(self, seq_nr, timeout=0):
        return self.sequencers[seq_nr].get_status()

    def get_acquisition_status(self, seq_nr, timeout=0):
        return self.sequencers[seq_nr].get_acquisition_status()

    def get_acquisitions(self, seq_nr: int, *, as_numpy: bool = False) -> dict:
        return self.sequencers[seq_nr].get_acquisition_data(as_numpy=as_numpy)

    def delete_acquisition_data(self, seq_nr, name='', all=False):
        self.sequencers[seq_nr].delete_acquisition_data(name=name, all=all)

    def start_adc_calib(self):
        if self._is_qrm:
            logger.info('Calibrate ADC')
        else:
            logger.error("QCM does not have method 'start_adc_calib'")

    def config_seq(self, seq_nr, name, value):
        self.sequencers[seq_nr].config(name, value)

    def config(self, name, value):
        for seq in self.sequencers:
            seq.config(name, value)

    def get_simulation_end_time(self):
        return max(seq.get_simulation_end_time() for seq in self.sequencers)

    def plot(self,
             t_min: float = None,
             t_max: float = None,
             channels: list[str] | list[int] | None = None,
             analogue_filter: bool = False,
             analogue_output_frequency: float = 4e9,
             **kwargs):
        for i, seq in enumerate(self.sequencers):
            if channels is not None and i not in channels and seq.label not in channels:
                # skip channel
                continue
            # assume only sequencers in sync mode have executed.
            if seq.sync_en():
                seq.plot(t_min=t_min, t_max=t_max,
                         analogue_filter=analogue_filter,
                         analogue_output_frequency=analogue_output_frequency)

    def get_output(self,
                   t_min: float = None,
                   t_max: float = None,
                   channels: list[str] | list[int] | None = None,
                   analogue_filter: bool = False,
                   output_frequency: float = 4e9,
                   ):
        output = {}
        for i, seq in enumerate(self.sequencers):
            if channels is not None and i not in channels and seq.label not in channels:
                # skip channel
                continue
            # assume only sequencers in sync mode have executed.
            if seq.sync_en():
                # sequencer may generate output on muliple outputs (I, Q, marker)
                seq_output = seq.get_output(t_min=t_min, t_max=t_max,
                                            analogue_filter=analogue_filter,
                                            output_frequency=output_frequency)
                output.update(seq_output)

        return output

    def print_acquisitions(self):
        for i, seq in enumerate(self.sequencers):
            data = self.get_acquisitions(i)
            if not len(data):
                continue
            for name, datadict in data.items():
                print(f"Acquisitions '{seq.name}':'{name}'")
                bins = datadict['acquisition']['bins']

                print("  'path0': [",
                      np.array2string(np.array(bins['integration']['path0']),
                                      prefix=' '*12,
                                      separator=',',
                                      threshold=100),
                      ']')
                print("  'path1': [",
                      np.array2string(np.array(bins['integration']['path1']),
                                      prefix=' '*12,
                                      separator=',',
                                      threshold=100),
                      ']')
                print("  'avg_cnt': [",
                      np.array2string(np.array(bins['avg_cnt']),
                                      prefix=' '*12,
                                      separator=',',
                                      threshold=100),
                      ']')

    def print_registers(self, seq_nr, reg_nrs=None):
        self.sequencers[seq_nr].print_registers(reg_nrs)


class Q1Simulator(qc.Instrument, Q1Module):
    _pulsar_parameters = [
        'reference_source',
        ]
    _log_only_params = [
        'led_brightness',
        ]

    def __init__(self, name, n_sequencers=6, sim_type=None):
        check_qblox_instrument_version()
        super().__init__(name)
        super().init_module(n_sequencers, sim_type)

        for par_name in self._pulsar_parameters:
            self.add_parameter(par_name, set_cmd=partial(self._set, par_name))

        for par_name in self._log_only_params:
            self.add_parameter(par_name,
                               set_cmd=partial(self._log_set, par_name))
        self.ignore_triggers = False

    def get_idn(self):
        return dict(vendor='Q1Simulator', model=self._sim_type, serial='', firmware='')

    @property
    def instrument_class(self):
        return InstrumentClass.PULSAR

    @property
    def instrument_type(self):
        return InstrumentType[self._sim_type]

    def reset(self):
        self.invalidate_cache()
        super().reset()

    def get_system_status(self):
        return SystemStatus(
            SystemStatuses.OKAY,
            [],
            SystemStatusSlotFlags({}))

    if qblox_version < Version('0.14'):
        # Deprecated API
        def get_system_state(self):
            return SystemState(
                SystemStatusOld.OKAY,
                [],
                SystemStatusSlotFlags({}))

    def start_sequencer(self, sequencer: int | None = None):
        if sequencer is not None:
            self.sequencers[sequencer].run()
            return

        # Get list of armed sequencers
        # pass to sequence executor

        sequencers = [self.sequencers[seq_number] for seq_number in self.armed_seq]
        run_sequencers(sequencers, self.ignore_triggers)

    def _log_set(self, name, value):
        logger.info(f'{self.name}: {name}={value}')

    def plot(self,
             t_min: float = None,
             t_max: float = None,
             channels: list[str] | list[int] | None = None,
             analogue_filter: bool = False,
             analogue_output_frequency: float = 4e9,
             **kwargs):
        """Plots the simulated output of the module.

        Args:
            t_min: minimum time in the plot.
            t_max: maximum time in the plot.
            channels: If not None specifies the channels to plot by name or sequencer number.
            analogue_filter: plot result after applying (estimated) analog filter.
        """
        pt.figure()
        super().plot(t_min=t_min, t_max=t_max, channels=channels,
                     analogue_filter=analogue_filter,
                     analogue_output_frequency=analogue_output_frequency)
        pt.grid(True)
        pt.legend()
        pt.xlabel('[ns]')
        pt.ylabel('[V]')
        pt.show()


def run_sequencers(sequencers: list[Q1Sequencer], ignore_triggers=False):
    if ignore_triggers:
        for seq in sequencers:
            seq.run()
        return

    # sort on used triggers

    # TODO Refactor trigger distribution in runtime distribution.
    seq_infos = [get_seq_trigger_info(seq) for seq in sequencers]
    trigger_dist = TriggerDistributor()
    seq_infos = sort_sequencers(seq_infos)
    for seq_info in seq_infos:
        seq = seq_info.sequencer
        seq.set_trigger_events(trigger_dist.get_trigger_events())
        seq.run()
        trigger_dist.add_emitted_triggers(seq.get_acq_trigger_events())



import re
from dataclasses import dataclass
from typing import Any

import numpy as np


@dataclass
class Instruction:
    text_line_nr: int
    mnemonic: str
    arglist: tuple[str] | None = None
    label: str | None = None
    args: list[int | str] | None = None
    reg_args: list[int] | None = None
    func_name: str | None = None
    func: Any = None
    clock_ticks: int = 1


class AsmSyntaxError(Exception):
    def __str__(self):
        # lines = [f"{self.__class__.__name__}:"] + list(self.args)
        return "\n\t" + "\n\t".join(self.args)


# S = immediate signed, U = immediate unsigned, R = register, L = label, D = destination register
mnemonic_args = {
    "illegal": "",
    "stop": "",
    "nop": "",
    "jmp": "URL",
    "jlt": "R,U,URL",
    "jge": "R,U,URL",
    "loop": "D,URL",
    "move": "URL,D",
    "not": "UR,D",
    "add": "R,UR,D",
    "sub": "R,UR,D",
    "and": "R,UR,D",
    "or": "R,UR,D",
    "xor": "R,UR,D",
    "asl": "R,UR,D",
    "asr": "R,UR,D",
    "set_mrk": "UR",
    "reset_ph": "",
    "set_freq": "SR",
    "set_ph": "UR",
    "set_ph_delta": "UR",
    "set_awg_gain": "SR,SR",
    "set_awg_offs": "SR,SR",
    "set_cond": "UR,UR,UR,U",
    "upd_param": "U",
    "play": "UR,UR,U",
    "acquire": "U,UR,U",
    "acquire_weighed": "U,UR,UR,UR,U",
    "set_latch_en": "UR,U",
    "latch_rst": "UR",
    "wait": "UR",
    "wait_sync": "UR",
    "wait_trigger": "UR",
}


class Q1Parser:

    def __init__(self):
        self.labels = {}

    def parse(self, program):
        labels = {}
        lines = program.split("\n")
        self.lines = lines

        instructions = []
        for i, line in enumerate(lines):
            label, mnemonic, arglist = self._parseline(line)
            if label:
                icnt = len(instructions)
                labels[label] = icnt
            if not mnemonic:
                continue

            instructions.append(Instruction(i, mnemonic, arglist, label))

        self.labels = labels

        for instr in instructions:
            mnemonic = instr.mnemonic
            func_name = "_" + mnemonic
            instr.func_name = func_name
            if mnemonic in mnemonic_args:
                try:
                    args, reg_args = self._evaluate_args(mnemonic_args[mnemonic], instr.arglist)
                    instr.args = args
                    instr.reg_args = reg_args
                    if reg_args and mnemonic not in ["jmp", "jge", "jlt", "loop"]:
                        # 1 cycle for every register. instr and 1st register are loaded in 1 cycle
                        instr.clock_ticks = len(reg_args)
                    else:
                        # 1 cycle to load instruction
                        instr.clock_ticks = 1
                except AsmSyntaxError as ex:
                    print(ex)
                    print(lines[instr.text_line_nr])
                    ex.args = ex.args + (f"line {instr.text_line_nr}: {lines[instr.text_line_nr]}", )
                    raise
                except Exception as ex:
                    print(ex)
                    print(lines[instr.text_line_nr])
                    ex.args = ex.args + (f"line {instr.text_line_nr}: {lines[instr.text_line_nr]}", )
                    raise
            else:
                instr.args = instr.arglist

        return lines, instructions

    def _parseline(self, line):
        org_line = line
        label_pattern = r"(\w+:)"
        instr_pattern = r"(\w+:)?\s*(\w+)\s*(.*)"
        if line.startswith("#Q1Sim:"):
            return self._parse_simcmd(line[7:])
        try:
            end = line.index("#")
            line = line[:end]
        except Exception:
            pass
        line = line.strip()
        if len(line) == 0:
            return [None, None, None]

        match = re.fullmatch(label_pattern, line)
        if match:
            label = match.group(1)
            label = label[:-1]
            return [label, None, None]

        match = re.fullmatch(instr_pattern, line)
        if match:
            label = match.group(1)
            if label:
                label = label[:-1]
            args = match.group(3).strip()
            if args:
                arglist = args.replace(" ", "").split(",")
            else:
                arglist = []
            return [label, match.group(2), arglist]
        raise Exception(f"{self.name}: Parse error on line: {org_line}")

    def _parse_simcmd(self, command):
        command = command.strip()
        # format: log "msg",register,options
        log_pattern = r'log "(.*)",(\w+)?,(\w+)?'
        match = re.fullmatch(log_pattern, command)
        if match:
            msg = match.group(1)
            register = match.group(2)
            options = match.group(3)
            if msg is None:
                msg = ""
            return None, "log", (msg, register, options)
        trigger_pattern = r"sim_trigger (\d),\s*([01])"
        match = re.fullmatch(trigger_pattern, command)
        if match:
            addr = match.group(1)
            value = match.group(2)
            return None, "sim_trigger", (addr, value)
        print(f"Unknown simulator command:{command}")
        return None, None, None

    def _evaluate_args(self, arg_types, args):
        types = arg_types.split(",") if arg_types else []
        args = list(args)
        if len(args) != len(types):
            raise AsmSyntaxError(f"Incorrect number of arguments {len(args)}<>{len(types)}")
        select_imm = False
        allow_imm = True
        reg_args = []
        for i, arg in enumerate(args):
            allowed = types[i]
            c = arg[0]
            if allowed == "D":
                if c != "R":
                    raise AsmSyntaxError("Destination must be register")
                args[i] = self._parse_reg_arg(arg)
            elif c == "@":
                if "L" not in allowed:
                    raise AsmSyntaxError(f"Label operand not support as argument {i}")
                if "R" in allowed:
                    select_imm = True
                args[i] = self._parse_label_arg(arg)
            elif c == "R":
                if "R" not in allowed:
                    raise AsmSyntaxError(f"Register operand not support as argument {i}")
                if "U" in allowed or "S" in allowed:
                    allow_imm = False
                args[i] = self._parse_reg_arg(arg)
                reg_args.append(i)
            else:
                if "R" in allowed:
                    select_imm = True
                if "U" in allowed:
                    args[i] = self._parse_uint32_arg(arg)
                elif "S" in allowed:
                    args[i] = self._parse_int32_arg(arg)
                else:
                    raise AsmSyntaxError(f"Immediate operand not support as argument {i}")

        if not allow_imm and select_imm:
            raise AsmSyntaxError("Combination of operands not supported")

        if len(reg_args) == 0:
            reg_args = None
        return args, reg_args

    def _parse_reg_arg(self, arg):
        try:
            reg_nr = int(arg[1:])
        except ValueError:
            raise AsmSyntaxError(f"Invalid register '{arg}'") from None
        if reg_nr < 0 or reg_nr > 63:
            raise AsmSyntaxError(f"Invalid register '{arg}'")
        return reg_nr

    def _parse_label_arg(self, arg):
        try:
            line_nr = self.labels[arg[1:]]
        except KeyError:
            raise AsmSyntaxError(f"Label {arg} not defined") from None
        return line_nr

    def _parse_uint32_arg(self, arg):
        try:
            res = np.uint32(arg)
        except (OverflowError, ValueError):
            raise AsmSyntaxError(f"Invalid unsigned integer: {arg}") from None
        return res

    def _parse_int32_arg(self, arg):
        try:
            res = np.int32(arg)
        except (OverflowError, ValueError):
            raise AsmSyntaxError(f"Invalid integer: {arg}")
        return res
